from typing import Iterable, Dict, Callable

import torch as t
import torch.distributions as d

from evostrat import Individual, Population


class CategoricalPopulation(Population):
    """
    A distribution over individuals whose parameters are sampled from categorical distributions
    """

    def __init__(self,
                 individual_parameter_shapes: Dict[str, t.Size],
                 individual_constructor: Callable[[Dict[str, t.Tensor]], Individual]
                 ):
        """
        A distribution over individuals whose parameters are sampled from categorical distributions

        The individuals are sampled by sampling their parameters from independent categorical distributions and then calling individual_constructor with the sampled parameters.

        :param individual_parameter_shapes: The shapes of the parameters of an individual. The last dimension denotes the number of classes of the categorical distribution.
        :param individual_constructor: A function that constructs an individual from parameters (with shapes equal to individual_parameter_shapes).
        """
        self.logits = {k: t.zeros(shape, requires_grad=True) for k, shape in individual_parameter_shapes.items()}
        self.constructor = individual_constructor

    def parameters(self) -> Iterable[t.Tensor]:
        return self.logits.values()

    def sample(self, n) -> Iterable[Individual]:
        samples = []
        for _ in range(n):
            classes = {k: d.Categorical(logits=logits).sample().detach() for k, logits in self.logits.items()}

            samples.append((
                self.constructor(classes),
                sum([d.Categorical(logits=logits).log_prob(classes[k]).sum() for k, logits in self.logits.items()])
            ))

        return samples



from typing import Iterable, Dict, Callable, Union

import torch as t
import torch.distributions as d

from evostrat import Individual, Population


class NormalPopulation(Population):
    """
    A distribution over individuals whose parameters are sampled from normal distributions
    """

    def __init__(self,
                 individual_parameter_shapes: Dict[str, t.Size],
                 individual_constructor: Callable[[Dict[str, t.Tensor]], Individual],
                 std: Union[float, str],
                 mirror_sampling: bool = True
                 ):
        """
        A distribution over individuals whose parameters are sampled from normal distributions

        The individuals are sampled by sampling their parameters from independent normal distributions and then calling individual_constructor with the sampled parameters.

        :param individual_parameter_shapes: The shapes of the parameters of an individual.
        :param individual_constructor: A function that constructs an individual from parameters (with shapes equal to individual_parameter_shapes).
        :param std: The standard deviation of the normal distributions.
        If a float, it is treated as a constant hyper-parameter. Equivalent to OpenAI ES [1].
        If it's a str it must be either 'shared' or 'diagonal':
            If it's 'shared' all parameters will share a single learned std deviation.
            if it's 'diagonal' each parameter will have their own learned std deviation, similar to PEPG [2].
        :param mirror_sampling: Whether or not individuals are sampled in pairs symmetrically around the mean. See [1].
        [1] - Salimans, Tim, et al. "Evolution strategies as a scalable alternative to reinforcement learning." arXiv preprint arXiv:1703.03864 (2017).
        [2] - Sehnke, Frank, et al. "Parameter-exploring policy gradients." Neural Networks 23.4 (2010): 551-559.
        """
        assert type(std) in {float, str}, "std must be a float or str"
        if type(std) == float:
            assert std > 0.0, "std must be greater than 0"
            self.param_logstds = {k: t.log(t.scalar_tensor(std)) for k in individual_parameter_shapes.keys()}
        if type(std) == str:
            assert std in {'shared', 'diagonal'}, "std must be 'shared' or 'diagonal'"
            if std == 'shared':
                self.shared_log_std = t.scalar_tensor(0.0, requires_grad=True)
                self.param_logstds = {k: self.shared_log_std for k in individual_parameter_shapes.keys()}
            else:
                self.param_logstds = {k: t.zeros(shape, requires_grad=True) for k, shape in individual_parameter_shapes.items()}

        self.std = std
        self.param_means = {k: t.zeros(shape, requires_grad=True) for k, shape in individual_parameter_shapes.items()}
        self.constructor = individual_constructor
        self.mirror_sampling = mirror_sampling

    def parameters(self) -> Iterable[t.Tensor]:
        if type(self.std) == float:
            std_params = []
        else:
            if self.std == 'shared':
                std_params = [self.shared_log_std]
            else:
                std_params = list(self.param_logstds.values())

        mean_params = list(self.param_means.values())
        return mean_params + std_params

    def sample(self, n) -> Iterable[Individual]:
        assert not self.mirror_sampling or n % 2 == 0, "if mirror_sampling is true, n must be an even number"

        n_samples = n // 2 if self.mirror_sampling else n
        samples = []
        for _ in range(n_samples):
            noise = {k: d.Normal(loc=t.zeros_like(v), scale=t.exp(self.param_logstds[k])).sample() for k, v in self.param_means.items()}
            samples.append((
                self.constructor({k: self.param_means[k] + n for k, n in noise.items()}),
                sum([d.Normal(self.param_means[k], scale=t.exp(self.param_logstds[k])).log_prob((self.param_means[k] + n).detach()).sum() for k, n in noise.items()])
            ))
            if self.mirror_sampling:
                samples.append((
                    self.constructor({k: self.param_means[k] - n for k, n in noise.items()}),
                    sum([d.Normal(self.param_means[k], scale=t.exp(self.param_logstds[k])).log_prob((self.param_means[k] - n).detach()).sum() for k, n in noise.items()])
                ))

        return samples



from .individual import Individual
from .population import Population
from .categorical_population import CategoricalPopulation
from .normal_population import NormalPopulation
from .util import compute_centered_ranks
from .util import normalize

__all__ = ["Population", "Individual", "CategoricalPopulation", "NormalPopulation", "compute_centered_ranks", "normalize"]



from typing import Dict

import gym
import torch as t
import tqdm as tqdm
from torch import nn
from torch.multiprocessing import Pool
from torch.optim import Adam

from evostrat import compute_centered_ranks, CategoricalPopulation, Individual


class BinaryLunarLander(Individual):
    """
    A lunar lander whose weights and biases are all either -1.0 or 1.0.
    """
    values = t.tensor([-1.0, 1.0], dtype=t.float32)

    def __init__(self):
        self.net = nn.Sequential(
            nn.Linear(8, 32), nn.Tanh(),
            nn.Linear(32, 4), nn.Softmax(dim=0)
        )

    @staticmethod
    def from_params(params: Dict[str, t.Tensor]) -> 'BinaryLunarLander':
        agent = BinaryLunarLander()
        params = {k: BinaryLunarLander.values[v] for k, v in params.items()}
        agent.net.load_state_dict(params)
        return agent

    def fitness(self, render=False) -> float:
        env = gym.make("LunarLander-v2")
        obs = env.reset()
        done = False
        r_tot = 0
        while not done:
            action = self.action(obs)
            obs, r, done, _ = env.step(action)
            r_tot += r
            if render:
                env.render()

        env.close()
        return r_tot

    def get_params(self) -> Dict[str, t.Tensor]:
        return self.net.state_dict()

    def action(self, obs):
        with t.no_grad():
            return t.argmax(self.net(t.tensor(obs, dtype=t.float32))).item()


if __name__ == '__main__':
    param_shapes = {k: v.shape + (2,) for k, v in BinaryLunarLander().get_params().items()}
    population = CategoricalPopulation(param_shapes, BinaryLunarLander.from_params)

    learning_rate = 0.1
    iterations = 1000
    pop_size = 200

    optim = Adam(population.parameters(), lr=learning_rate)
    pbar = tqdm.tqdm(range(iterations))
    pool = Pool()

    for _ in pbar:
        optim.zero_grad()
        raw_fit = population.fitness_grads(pop_size, pool, compute_centered_ranks)
        optim.step()
        pbar.set_description("fit avg: %0.3f, std: %0.3f" % (raw_fit.mean().item(), raw_fit.std().item()))
        if raw_fit.mean() > 200:
            print("Solved.")
            break

    pool.close()



from typing import Iterable
import torch as t


def compute_centered_ranks(x: Iterable[float]) -> Iterable[float]:
    """
    Compute centered ranks, e.g. [-81.0, 11.0, -0.5] --> [-0.5, 0.5, 0.0]
    """
    x = t.tensor(x)
    assert x.ndim == 1
    ranks = t.zeros((len(x),), dtype=t.long)
    ranks[x.argsort()] = t.arange(len(x))
    ranks = ranks.to(t.float32)
    ranks = ranks / (len(x) - 1)
    ranks = ranks - 0.5
    return ranks.tolist()


def normalize(x: Iterable[float]) -> Iterable[float]:
    """
    Normalize a list of floats to have zero mean and variance 1
    """
    x = t.tensor(x)
    assert x.ndim == 1
    x -= x.mean()
    x /= (x.std() + 1e-8)
    return x.tolist()
